from datetime import datetime

from mock import patch
import pytest

from blockerbugs import db
from blockerbugs.models.spin import Spin, SpinState, SpinType
from testing.test_controllers import add_release, add_milestone, \
    add_bug, add_spin


class TestSpinModel(object):
    @classmethod
    def setup_class(cls):
        cls.mocked_date = datetime(1990, 1, 1)
        patcher_config = {'utcnow.return_value': cls.mocked_date,
                          'side_effect': lambda *args, **kw: datetime(*args, **kw)}
        cls.patcher = patch('blockerbugs.models.spin.datetime', **patcher_config)
        cls.patcher.start()
        db.session.rollback()
        db.drop_all()
        db.create_all()
        cls.release = add_release(99)
        cls.milestone = add_milestone(cls.release, 'final', 100, 101,
                                       '99-final', True)

    @classmethod
    def teardown_class(cls):
        cls.patcher.stop()
        db.session.rollback()
        db.drop_all()

    def test_set_date_requested_default(self):
        spin = add_spin('test-spin', SpinType.TC, [],
                         self.milestone, state=SpinState.obsolete)
        assert spin.date_requested is not None

    def test_set_date_requested_then_state_requested(self):
        spin = add_spin('test-spin', SpinType.TC, [],
                        self.milestone, state=SpinState.obsolete)
        spin.date_requested = datetime(2001, 1, 1)
        spin.state = SpinState.requested
        assert spin.date_requested == self.mocked_date

    def test_set_date_created_is_none_and_state_built(self):
        spin = add_spin('test-spin', SpinType.TC, [],
                        self.milestone, state=SpinState.obsolete)
        assert spin.date_created is None
        spin.state = SpinState.built
        assert spin.date_created == self.mocked_date

    def test_state_built_and_date_created_is_not_null(self):
        spin = add_spin('test-spin', SpinType.TC, [],
                        self.milestone, state=SpinState.obsolete)
        dt = datetime(2001, 1, 1)
        spin.date_created = dt
        spin.state = SpinState.built
        assert spin.date_created == dt

    def test_spin_state_from_string(self):
        spin = add_spin('test-spin', SpinType.TC, [],
                        self.milestone, state=SpinState.obsolete)
        spin.state = 'built'
        assert SpinState.built == spin.state

    def test_spin_type_from_string(self):
        spin = add_spin('test-spin', SpinType.TC, [],
                        self.milestone, state=SpinState.obsolete)
        spin.spin_type = 'RC'
        assert spin.spin_type == SpinType.RC

    def test_state_from_bad_string(self):
        spin = add_spin('test-spin', SpinType.TC, [],
                        self.milestone, state=SpinState.obsolete)
        with pytest.raises(ValueError):
            spin.state = 'foo'


    def test_spin_type_from_bad_string(self):
        spin = add_spin('test-spin', SpinType.TC, [],
                        self.milestone, state=SpinState.obsolete)
        with pytest.raises(ValueError):
            spin.spin_type = 'foo'
