from itertools import chain
from datetime import datetime
from copy import copy

from bunch import Bunch
from mock import patch, Mock
from fedora.client import ServerError

from blockerbugs import db
from blockerbugs.models.milestone import Milestone
from blockerbugs.models.release import Release
from blockerbugs.models.update import Update
from blockerbugs.util.update_sync import UpdateSync
from testing.test_controllers import add_bug


# only required part of a Bodhi response
base_update = Bunch(
    title=u'anaconda-18.6.8-1.fc18',
    date_pushed=u'2012-09-13 16:40:28',
    date_submitted=u'2012-09-12 23:43:04',
    karma=1,
    request=u'testing',
    stable_karma=3,
    status=u'pending',
    bugs=[
        Bunch(bz_id=2000, parent=False, security=False,
              title=u'bodhi bug 2000 title'),
        Bunch(bz_id=2001, parent=False, security=False,
              title=u'bodhi bug 2001 title')],
    release=Bunch(dist_tag=u'f99', id_prefix=u'FEDORA', locked=True,
                  long_name=u'Fedora 99', name=u'F99'),
    updateid='FEDORA-2012-13902',
)

update_for_bugs_2000_2001 = copy(base_update)

update1_for_bug_3000 = copy(base_update)
update1_for_bug_3000.update({
    'title': u'libreport-2.1.3-2.fc19',
    'bugs': [Bunch(bz_id=3000, parent=False, security=False,
                   title=u"anaconda can't report traceback to bugzilla")],
})

update2_for_bug_3000 = copy(base_update)
update2_for_bug_3000.update({
    'title': u'jboss-servlet-2.5-api-1.0.1-3.fc19,resteasy-2.3.2-12.fc19',
    'bugs': [Bunch(bz_id=3000, parent=False, security=False,
                   title=u"anaconda can't report traceback to bugzilla")],
})


class FakeBodhiInterface(object):
    updates = {
        ('f99', '2000'): [update_for_bugs_2000_2001],
        ('f99', '2001'): [update_for_bugs_2000_2001],
        ('f99', '3000'): [update1_for_bug_3000, update2_for_bug_3000],
        ('f99', '4000'): [],
    }

    updates_time = {
        update_for_bugs_2000_2001.title: datetime(2001, 1, 1),
        update1_for_bug_3000.title: datetime(2002, 1, 1),
        update2_for_bug_3000.title: datetime(2003, 1, 1),
    }

    def query(self, release=None, bugs=None, **kwargs):
        updates = list(chain(*(self.updates[(release, bug)] for bug in bugs)))
        return Bunch(updates=updates,
                     page=1,
                     pages=1,
                     rows_per_page=20,
                     total=len(updates))

class FakeErrorBodhiInterface(object):
    def query(self, **kwargs):
        return Bunch(errors=[Bunch(location=u"querystring",
                                    name=u"release",
                                    description=u"Unknown release u'f99'")],
                     status=u"error")


class FakeRaisesBodhiInterface(object):
    def query(self, **kwargs):
        raise ServerError('https://admin.fedoraproject.org/updates/list',
                          500,
                          'Internal Server Error')


class TestfuncUpdateSync(object):
    def setup_method(self, method):
        db.session.flush()
        db.session.rollback()
        db.drop_all()
        db.create_all()
        self.test_release99 = Release(99)
        self.test_milestone99alpha = Milestone(
            self.test_release99, 'alpha', 1000, 1001, '99-final')
        self.test_milestone99alpha.active = True
        db.session.add(self.test_release99)
        db.session.add(self.test_milestone99alpha)
        db.session.commit()
        self.update_sync = UpdateSync(db, FakeBodhiInterface)

    def test_sync_single_bug_with_one_update(self):
        bug1 = add_bug(2000, 'testbug1', self.test_milestone99alpha)
        self.update_sync.sync_updates(self.test_release99)
        updates = Update.query.all()
        assert len(updates) == 1
        update = updates[0]
        assert len(update.bugs) == 1  # second bug not in db
        assert update.bugs[0].bugid == bug1.bugid
        assert update.status == u'testing'
        assert update.pending

    def test_sync_two_bugs_with_one_update(self):
        bug1 = add_bug(2000, 'testbug1', self.test_milestone99alpha)
        bug2 = add_bug(2001, 'testbug2', self.test_milestone99alpha)
        self.update_sync.sync_updates(self.test_release99)
        updates = Update.query.all()
        assert len(updates) == 1
        update = updates[0]
        assert len(update.bugs) == 2
        bugs_id = [bug.bugid for bug in update.bugs]
        assert bug1.bugid in bugs_id
        assert bug2.bugid in bugs_id

    def test_search_updates_two_bugs_one_update(self):
        bug1 = add_bug(2000, 'testbug1', self.test_milestone99alpha)
        bug2 = add_bug(2001, 'testbug2', self.test_milestone99alpha)
        updates = self.update_sync.search_updates([bug1.bugid, bug2.bugid],
                                                  self.test_release99.number)
        assert len(updates) == 1  # duplicates resolved in search
        update = updates[0]
        assert len(update['bugs']) == 2
        assert bug1.bugid in update['bugs']
        assert bug2.bugid in update['bugs']

    def test_sync_one_bug_with_two_updates(self):
        bug1 = add_bug(3000, 'testbug1', self.test_milestone99alpha)
        self.update_sync.sync_updates(self.test_release99)
        updates = Update.query.all()
        assert len(updates) == 2
        update1, update2 = updates
        assert update1.id != update2.id
        assert len(update1.bugs) == len(update2.bugs) == 1
        assert update1.bugs[0].bugid == update2.bugs[0].bugid == bug1.bugid

    def test_sync_multiple_bugs_with_multiple_updates(self):
        self.test_release99.last_update_sync = datetime(2020, 1, 1)
        bug1 = add_bug(2000, 'testbug1', self.test_milestone99alpha)
        bug2 = add_bug(2001, 'testbug2', self.test_milestone99alpha)
        bug3 = add_bug(3000, 'testbug3', self.test_milestone99alpha)
        self.update_sync.sync_updates(self.test_release99)
        updates = Update.query.all()
        assert len(updates) == 3
        updates_title = set(update.title for update in updates)
        assert updates_title == set((update_for_bugs_2000_2001.title,
                                     update1_for_bug_3000.title,
                                     update2_for_bug_3000.title))
        bugs_id = set(bug.bugid for update in updates for bug in update.bugs)
        assert bugs_id == set((bug1.bugid, bug2.bugid, bug3.bugid))

    def test_no_updates_for_bug(self):
        bug1 = add_bug(4000, 'testbug1', self.test_milestone99alpha)
        self.update_sync.sync_updates(self.test_release99)
        updates = Update.query.all()
        assert len(updates) == 0

    def test_sync_release_no_bugs(self):
        with patch('blockerbugs.util.update_sync.datetime') as mock_date:
            mock_date.utcnow.return_value = datetime(2000, 1, 1)
            mock_date.side_effect = lambda *args, **kw: datetime(*args, **kw)
            self.update_sync.sync_updates(self.test_release99)
            assert self.test_release99.last_update_sync == datetime(2000, 1, 1)
            updates = Update.query.all()
            assert len(updates) == 0

    def test_release_not_found_error(self):
        add_bug(4000, 'testbug1', self.test_milestone99alpha)
        update_sync = UpdateSync(db, FakeErrorBodhiInterface)
        log_error = Mock(wraps=update_sync.log.error)
        update_sync.log.error = log_error
        update_sync.sync_updates(self.test_release99)
        assert log_error.call_count == 1
        assert 'Unknown release' in log_error.call_args[0][0]

    def test_release_server_error(self):
        add_bug(4000, 'testbug1', self.test_milestone99alpha)
        update_sync = UpdateSync(db, FakeRaisesBodhiInterface)
        log_error = Mock(wraps=update_sync.log.error)
        update_sync.log.error = log_error
        update_sync.sync_updates(self.test_release99)
        assert log_error.call_count == 1
        assert 'Internal Server Error' in log_error.call_args[0][0]

    def test_sync_clean_updates(self):
        add_bug(3000, 'testbug1', self.test_milestone99alpha)

        self.update_sync.sync_updates(self.test_release99)
        updates = Update.query.filter(Update.status == 'deleted').all()
        assert len(updates) == 0

        b = FakeBodhiInterface
        b.updates[('f99', '3000')] = [update1_for_bug_3000]
        update_sync = UpdateSync(db, b)

        update_sync.sync_updates(self.test_release99)
        updates = Update.query.filter(Update.status == 'deleted').all()
        assert len(updates) == 1
